/*
    CustomMetrics - Simple, configurable, economical metrics for AWS

    See the README for the Metric schema details.
 */
import process from 'process';
import { DynamoDBClient, GetItemCommand, PutItemCommand, QueryCommand, } from '@aws-sdk/client-dynamodb';
import { marshall, unmarshall } from '@aws-sdk/util-dynamodb';
const Version = 1;
//MOB disable
const Assert = true; // Enable asserts
const Buffering = true; // Enable buffered metrics
const DefaultResolution = 0; // Default number of p-values to store
const MaxSeq = Number.MAX_SAFE_INTEGER; // Maximum sequence number for collision detection
const MaxRetries = 10; // Max retries when emitting a metric and encountering collisions
const MetricListLimit = 10000; // Sanity limit for getting a list of metrics
/*
    Default spans are configurable by the constructor
 */
export const DefaultSpans = [
    { period: 5 * 60, samples: 10 }, //  300, 5 mins, interval: 30 secs
    { period: 60 * 60, samples: 12 }, //  3600, 1 hr, interval: 5 mins
    { period: 24 * 60 * 60, samples: 12 }, // 86400, 24 hrs, interval: 2 hrs
    { period: 7 * 24 * 60 * 60, samples: 14 }, // 604,800, 7 days, interval: 1/2 day
    { period: 28 * 24 * 60 * 60, samples: 14 }, // 2,419,200, 28 days, interval: 2 days
    { period: 365 * 24 * 60 * 60, samples: 12 }, // 31,536,000, 1 year, interval: 1 month
];
var Instances = {};
/*
    On exit, flush any buffered metrics. This requires any Lambda Layer to receive this signal
 */
process.on('SIGTERM', 
/* istanbul ignore next */
async () => {
    /* istanbul ignore next */
    await CustomMetrics.terminate();
});
export class CustomMetrics {
    consistent = false;
    buffer;
    buffers = null;
    client;
    expires;
    log;
    options;
    owner;
    prefix = 'metric';
    primaryKey;
    sortKey;
    pResolution;
    source;
    spans;
    table;
    type;
    ttl;
    constructor(options = {}) {
        this.log = new Log(options.log);
        if (options.ttl && typeof options.ttl != 'number') {
            throw new Error('Bad type for "ttl" option');
        }
        if (options.spans && (!Array.isArray(options.spans) || options.spans.length == 0)) {
            throw new Error('The "spans" option must be an non-empty array');
        }
        if (options.source && typeof options.source != 'string') {
            throw new Error('Non-string "source" option');
        }
        if (options.pResolution != undefined && (options.pResolution < 0 || options.pResolution > 1000)) {
            throw new Error('Invalid "pResolution" option. Must be between 0 and 1000. Default is 0');
        }
        if (options.consistent != null && typeof options.consistent != 'boolean') {
            throw new Error('Bad type for "consistent" option');
        }
        if (options.prefix) {
            this.prefix = options.prefix;
        }
        if (options.buffer) {
            if (typeof options.buffer != 'object') {
                throw new Error('Bad type for "buffer" option');
            }
            this.buffer = options.buffer;
        }
        this.expires = options.expires || 'expires';
        this.primaryKey = options.primaryKey || 'pk';
        this.sortKey = options.sortKey || 'sk';
        this.type = options.type || { _type: 'Metric' };
        /* istanbul ignore else */
        if (options.client) {
            this.client = options.client;
        }
        else {
            let params = {};
            if (options.creds) {
                params.credentials = options.creds;
                //  Allow region in credentials
                params.region = params.credentials.region;
            }
            if (options.region) {
                params.region = options.region;
            }
            this.client = new DynamoDBClient(params);
        }
        if (!options.table) {
            throw new Error('Missing DynamoDB table name property');
        }
        /* istanbul ignore next */
        this.table = options.table;
        this.options = options;
        this.owner = options.owner || 'default';
        this.spans = options.spans || DefaultSpans;
        this.ttl = options.ttl || this.spans[this.spans.length - 1].period;
        if (options.consistent != null) {
            this.consistent = options.consistent;
        }
        if (options.source) {
            this.source = options.source;
        }
        this.pResolution = options.pResolution || DefaultResolution;
    }
    async emit(namespace, metricName, value, dimensionsList = [{}], options = {}) {
        if (value == undefined || value == null) {
            throw new Error('Invalid metric value');
        }
        if (dimensionsList.length == 0) {
            dimensionsList = [{}];
        }
        value = Number(value);
        if (isNaN(value)) {
            throw new Error(`Value to emit is not valid`);
        }
        if (!namespace || !metricName) {
            throw new Error('Missing emit namespace / metric argument');
        }
        /* istanbul ignore next */
        if (!Array.isArray(dimensionsList)) {
            throw new Error('Dimensions must be an array');
        }
        if (dimensionsList.length == 0) {
            dimensionsList = [{}];
        }
        let point;
        point = { count: 1, sum: value };
        return await this.emitDimensions(namespace, metricName, point, dimensionsList, options);
    }
    async emitDimensions(namespace, metricName, point, dimensionsList, options) {
        let result;
        for (let dim of dimensionsList) {
            let dimensions = this.makeDimensionString(dim);
            let buffer = options.buffer || this.buffer;
            if (buffer && (buffer.elapsed || buffer.force || buffer.sum || buffer.count) && Buffering) {
                result = await this.bufferMetric(namespace, metricName, point, dimensions, options);
            }
            else {
                result = await this.emitDimensionedMetric(namespace, metricName, point, dimensions, options);
            }
        }
        return result;
    }
    /*
        Buffer a metric for specific dimensions
        Point values are uniquely buffered in elements indexed by a (namespace, metric, dimensions) key.
        Buffered values are flushed when a sum, count or timespan parameter is exceeded.
        The accumulated buffered values are returned as a metric result until the buffered values are
        flushed. These values will be only the buffered values for this Lambda.
        When the metric is flushed, the full persisted metric is returned with all spans.
        This means returned metrics will typically be just accumulated buffered values and won't reflect
        other lambdas until flushed. If you need consistent return values -- use query().
     */
    async bufferMetric(namespace, metricName, point, dimensions, options) {
        let buffer = options.buffer || this.buffer;
        let key = this.getBufferKey(namespace, metricName, dimensions);
        let buffers = (this.buffers = this.buffers || {});
        let now = Math.floor((options.timestamp || Date.now()) / 1000);
        let elapsed = buffer.elapsed || this.spans[0].period / this.spans[0].samples;
        let elt = (buffers[key] = buffers[key] || {
            count: 0,
            sum: 0,
            timestamp: now + elapsed,
            elapsed: elapsed,
            namespace: namespace,
            metric: metricName,
            dimensions,
            spans: [{ points: [{ count: 0, sum: 0 }] }],
        });
        /*
            Add point value to the lowest span and to the elt (to manage when to persist)
         */
        let current = elt.spans[0].points.at(-1);
        if (current) {
            current.count += point.count;
            current.sum += point.sum;
        }
        elt.count += point.count;
        elt.sum += point.sum;
        if (buffer.force ||
            (buffer.sum && elt.sum >= buffer.sum) ||
            (buffer.count && elt.count >= buffer.count) ||
            now >= elt.timestamp) {
            options = Object.assign({}, options, { timestamp: now * 1000 });
            let metric = await this.emitDimensionedMetric(namespace, metricName, elt, dimensions, options);
            //  Reset tallies and save higher spans to return for future buffered metrics
            elt.count = elt.sum = 0;
            elt.spans = metric.spans;
            elt.timestamp = now + (buffer.elapsed || this.spans[0].period / this.spans[0].samples);
            return metric;
        }
        CustomMetrics.saveInstance({ key }, this);
        return {
            spans: elt.spans,
            metric: metricName,
            namespace: namespace,
            owner: options.owner || this.owner,
            version: Version,
        };
    }
    /*
        Emit a metric for specific dimensions
     */
    async emitDimensionedMetric(namespace, metricName, point, dimensions, options = {}) {
        /*
            Update the metric. May need retries if colliding with other updaters.
         */
        let now = Math.floor((options.timestamp || Date.now()) / 1000);
        let ttl = options.ttl != undefined ? options.ttl : this.ttl;
        let retries = MaxRetries;
        let metric;
        let backoff = 10;
        /* istanbul ignore next */
        let chan = options.log == true ? 'info' : 'trace';
        do {
            let owner = options.owner || this.owner;
            metric = await this.getMetric(owner, namespace, metricName, dimensions, options.log);
            if (metric) {
                if (options.upgrade) {
                    metric = this.upgradeMetric(metric);
                }
            }
            else {
                metric = this.initMetric(owner, namespace, metricName, dimensions, now);
            }
            if (point.timestamp) {
                /*
                    Buffered points only.
                    Pick the first span for which the point is after the earliest span start or after the end of the span
                 */
                let si = metric.spans.findIndex((s) => s.end - s.period <= point.timestamp || s.end <= point.timestamp);
                /* istanbul ignore else */
                if (si >= 0) {
                    this.addValue(metric, point.timestamp, point, si);
                }
                else {
                    //  Silently discard
                }
            }
            else {
                this.addValue(metric, now, point, 0);
            }
            /* istanbul ignore next */
            if (this.source) {
                metric._source = this.source;
            }
            /*
                Set the expiration TTL. Defaults to the longest span.
                Users may define a shorter TTL to prune metrics for inactive items.
            */
            if (ttl) {
                //  MOB - is this a date, seconds or msec
                metric.expires = now + ttl;
            }
            if (await this.putMetric(metric, options)) {
                break;
            }
            /* istanbul ignore next */
            if (retries == 0) {
                this.log.error(`Metric update has too many retries`, { namespace, metricName, dimensions });
                break;
            }
            /* istanbul ignore next */
            this.log[chan](`Retry ${MaxRetries - retries} metric update ${metric.namespace} ${metric.metric} ${metric.dimensions}`, {
                retries,
                metric,
            });
            //  Exponential backoff
            /* istanbul ignore next */
            backoff = backoff * 2;
            /* istanbul ignore next */
            this.log[chan](`Retry backoff ${backoff} ${this.jitter(backoff)}`);
            /* istanbul ignore next */
            await this.delay(this.jitter(backoff));
        } while (retries-- > 0);
        return metric;
    }
    /*
        Upgrade a metric with new spans. Only the specified dimensions are upgraded.
     */
    async upgrade(namespace, metricName, dimensionsList = [{}], options = {}) {
        let owner = options.owner || this.owner;
        if (dimensionsList.length == 0) {
            dimensionsList = [{}];
        }
        let metric;
        for (let dim of dimensionsList) {
            let dimensions = this.makeDimensionString(dim);
            let old = await this.getMetric(owner, namespace, metricName, dimensions, options.log);
            metric = this.upgradeMetric(old);
            await this.putMetric(metric, options);
        }
        return metric;
    }
    /*
        Upgrade a metric and return the upgraded result
        Optimized for when an upgrade is not required.
     */
    upgradeMetric(old) {
        let required = false;
        /*
            Check if upgrade required
         */
        if (this.spans.length == old.spans.length) {
            for (let [index, span] of Object.entries(old.spans)) {
                if (span.period != this.spans[index].period || span.samples != this.spans[index].samples) {
                    required = true;
                }
            }
            if (!required) {
                return old;
            }
        }
        /*
            This initializes a new metric with the new spans and apportion the old data points to
            the new metric at the point's timestamp. Pick the earliest timestamp from the old metric.
         */
        let now = Math.min(...old.spans.map((span) => span.end - span.period)) || Math.floor(Date.now() / 1000);
        let metric = this.initMetric(old.owner, old.namespace, old.metric, old.dimensions, now);
        for (let span of old.spans) {
            let interval = span.period / span.samples;
            let firstPoint = span.end - span.points.length * interval;
            /*
                Pick the first span for which the point is after the earliest span start or after the end of the span
            */
            let si = metric.spans.findIndex((s) => s.end - s.period <= firstPoint || s.end <= firstPoint);
            let t = firstPoint;
            for (let point of span.points) {
                this.addValue(metric, t, point, si);
                t += interval;
            }
        }
        return metric;
    }
    /*
        Flush metrics for all instances on Lambda termination
     */
    static async terminate() {
        await CustomMetrics.flushAll();
    }
    static async flushAll() {
        for (let [key, instance] of Object.entries(Instances)) {
            await instance.flush();
            CustomMetrics.freeInstanceByKey(key);
        }
        Instances = {};
    }
    async flush(options = {}) {
        if (!this.buffers)
            return;
        let now = Math.floor((options.timestamp || Date.now()) / 1000);
        for (let elt of Object.values(this.buffers)) {
            await this.flushElt(elt, now);
        }
    }
    async flushElt(elt, now) {
        //  Choose timestamp if before the buffer expires, otherwise choose the buffer expiry time
        elt.timestamp = Math.min(now, elt.timestamp);
        let metric = await this.emitDimensionedMetric(elt.namespace, elt.metric, elt, elt.dimensions, {
            timestamp: elt.timestamp * 1000,
        });
        elt.count = elt.sum = 0;
        elt.spans = metric.spans;
        elt.timestamp = now + (elt.elapsed || this.spans[0].period / this.spans[0].samples);
    }
    getBufferKey(namespace, metricName, dimensions) {
        return `${namespace}|${metricName}|${JSON.stringify(dimensions)}`;
    }
    /*
        Query metrics. Return an array of metrics
     */
    async query(namespace, metricName, dimensions, period, statistic, options = {}) {
        let owner = options.owner || this.owner;
        let dimString = this.makeDimensionString(dimensions);
        if (period > this.spans.at(-1).period) {
            period = this.spans.at(-1).period;
        }
        let now = Math.floor((options.timestamp || Date.now()) / 1000);
        /*
           Flush buffered metrics for this instance. Will not see buffered metrics in other instances
           until they are flushed.
         */
        if (this.buffers) {
            let key = this.getBufferKey(namespace, metricName, dimString);
            if (this.buffers[key]) {
                await this.flushElt(this.buffers[key], now);
            }
        }
        let metric = await this.getMetric(owner, namespace, metricName, dimString, options.log);
        if (!metric) {
            return { dimensions, id: options.id, metric: metricName, namespace, period, points: [], owner, samples: 0 };
        }
        let result = this.processMetric(metric, now, period, statistic, options);
        /* istanbul ignore next */
        this.log[options.log == true ? 'info' : 'trace'](`Query metrics ${namespace}, ${metricName}`, {
            dimensions,
            period,
            statistic,
            options,
            result,
        });
        return result;
    }
    async queryMetrics(namespace, metric, period, statistic, options = {}) {
        let owner = options.owner || this.owner;
        let next = options.next;
        let limit = options.limit || MetricListLimit;
        /* istanbul ignore next */
        let chan = options.log == true ? 'info' : 'trace';
        let items, command;
        let count = 0;
        do {
            /* istanbul ignore next */
            ;
            ({ command, items, next } = await this.findMetrics(owner, namespace, metric, limit, next, 'spans'));
            this.log[chan](`Find metrics ${namespace}, ${metric}`, { command, items });
            if (items.length) {
                count += items.length;
            }
        } while (next && count < limit);
        let now = Math.floor((options.timestamp || Date.now()) / 1000);
        let results = [];
        for (let metric of items) {
            let result = this.processMetric(metric, now, period, statistic, { accumulate: true, timestamp: now });
            results.push(result);
        }
        return results;
    }
    /*
       Process a metric for query() or queryMetrics() and extract the desired series or accumlated value
     */
    processMetric(metric, now, period, statistic, options) {
        let end;
        let si;
        let owner = options.owner || this.owner;
        if (options.start) {
            /*
                Find span for the request start: period < span[i+1].period
            */
            let start = options.start / 1000;
            si = metric.spans.findIndex((s) => period <= s.period && s.end - s.period <= start && start < s.end);
            end = start + period;
        }
        else {
            let span = metric.spans[0];
            let interval = span.period / span.samples;
            /*
                Special case: if now is within the most recent span interval, then set "start" to include this interval.
            */
            if (span.end - interval <= now && now < span.end) {
                end = span.end;
            }
            else {
                end = now;
            }
            si = metric.spans.findIndex((s) => period <= s.period);
        }
        if (si < 0) {
            si = metric.spans.length - 1;
        }
        /*
            Aggregate data for all spans up to the desired span. Do this because spans are updated lazily on emit.
         */
        this.addValue(metric, now, { count: 0, sum: 0 }, 0, si);
        let span = metric.spans[si];
        let result;
        if (options.accumulate) {
            result = this.accumulateMetric(metric, span, statistic, owner, end, period);
        }
        else {
            result = this.calculateSeries(metric, span, statistic, owner, end, period);
        }
        result.id = options.id;
        return result;
    }
    /*
        Accumulate the metric data points into a single value.
        This is useful for gauges and widgets that need a single data value for the metric.
        According to the desired statistic, this calculates the avg, count, max, min, sum, and pvalues.
     */
    accumulateMetric(metric, span, statistic, owner, end, period) {
        let start = this.alignTime(span, end - period);
        let value = 0, count = 0, pvalues = [];
        if (statistic == 'max') {
            value = Number.NEGATIVE_INFINITY;
        }
        else if (statistic == 'min') {
            value = Infinity;
        }
        else if (statistic == 'sum') {
            value = 0;
            count = 0;
        }
        else if (statistic == 'count') {
            value = 0;
            count = 0;
        }
        else if (statistic == 'current') {
            value = 0;
            count = 0;
        }
        else if (statistic.match(/^p[0-9]+/)) {
            pvalues = [];
        } /* avg */
        else {
            value = 0;
            count = 0;
        }
        let points = span.points;
        let interval = span.period / span.samples;
        let t = span.end - span.points.length * interval;
        for (let i = 0; i < points.length; i++) {
            let point = points[i];
            if (start <= t && t < start + period) {
                if (statistic == 'max') {
                    if (point.max != undefined) {
                        value = Math.max(value, point.max);
                    }
                    else {
                        //  For use to accumulate AWS metrics that don't keep min/max in points
                        value = Math.max(value, point.sum / (point.count || 1));
                    }
                }
                else if (statistic == 'min') {
                    if (point.min != undefined) {
                        value = Math.min(value, point.min);
                    }
                    else {
                        //  For use to accumulate AWS metrics that don't keep min/max in points
                        value = Math.min(value, point.sum / (point.count || 1));
                    }
                }
                else if (statistic == 'sum') {
                    value += point.sum;
                }
                else if (statistic == 'current') {
                    value = point.sum / (point.count || 1);
                }
                else if (statistic == 'count') {
                    value += point.count;
                }
                else if (statistic.match(/^p[0-9]+/)) {
                    pvalues = pvalues.concat(point.pvalues);
                } /* avg */
                else {
                    value += point.sum;
                }
                count += point.count;
            }
            t += interval;
        }
        if (statistic.match(/^p[0-9]+/)) {
            let p = parseInt(statistic.slice(1));
            pvalues.sort((a, b) => a - b);
            let nth = Math.min(Math.round((pvalues.length * p) / 100 + 1), pvalues.length - 1);
            value = pvalues[nth];
        }
        else if (statistic == 'avg') {
            value /= Math.max(count, 1);
        }
        return {
            dimensions: this.makeDimensionObject(metric.dimensions),
            metric: metric.metric,
            namespace: metric.namespace,
            owner: owner,
            period: span.period,
            points: [{ value, timestamp: start + period, count }],
            samples: span.samples,
        };
    }
    /*
        Process the metric points. This is used for graphs that need all the data points.
        This calculates the avg, max, min, sum, count, pvalues and per-point timestamps.
        This always returns a full series of points even if there is no data.
     */
    calculateSeries(metric, span, statistic, owner, end, period) {
        let points = [];
        let interval = span.period / span.samples;
        /*
            Start points aligned with span buckets
         */
        let start = this.alignTime(span, end - period);
        let firstPoint = span.end - span.points.length * interval;
        let t;
        for (t = start; t < firstPoint && points.length < span.samples; t += interval) {
            points.push({ value: 0, count: 0, timestamp: t * 1000 });
        }
        t = firstPoint;
        for (let point of span.points) {
            //  MOB - end was start + period, now end is closer
            if (start <= t && t < end) {
                let value = undefined;
                /* istanbul ignore else */
                if (point.count > 0) {
                    if (statistic == 'max') {
                        if (point.max != undefined) {
                            if (value == undefined) {
                                value = point.max;
                            }
                            else {
                                value = Math.max(value, point.max);
                            }
                        }
                    }
                    else if (statistic == 'min') {
                        if (point.min != undefined) {
                            if (value == undefined) {
                                value = point.min;
                            }
                            else {
                                value = Math.min(value, point.min);
                            }
                        }
                    }
                    else if (statistic == 'sum') {
                        value = point.sum;
                    }
                    else if (statistic == 'count') {
                        value = point.count;
                    }
                    else if (statistic.match(/^p[0-9]+/)) {
                        let p = parseInt(statistic.slice(1));
                        let pvalues = point.pvalues;
                        pvalues.sort((a, b) => a - b);
                        let nth = Math.min(Math.round((pvalues.length * p) / 100 + 1), pvalues.length - 1);
                        value = pvalues[nth];
                    } /* avg | current */
                    else {
                        value = point.sum / point.count;
                    }
                }
                else {
                    value = 0;
                }
                /*
                    The timestamp is set to be the end of the point bucket not the start, and not after the range
                */
                let timestamp = Math.min(t + interval, end) * 1000;
                points.push({ value, count: point.count, timestamp });
            }
            t += interval;
        }
        let count = Math.min(Math.ceil(period / interval), span.samples);
        while (points.length < count) {
            let timestamp = Math.min(t + interval, end) * 1000;
            points.push({ value: 0, count: 0, timestamp });
            t += interval;
        }
        return {
            dimensions: this.makeDimensionObject(metric.dimensions),
            metric: metric.metric,
            namespace: metric.namespace,
            period: span.period,
            points: points,
            owner: owner,
            samples: span.samples,
        };
    }
    /*
        Convert a dimensions object of the form {key: value, ...} to a string with comma separated "key=value,..." with sorted property keys
     */
    makeDimensionString(dimensions) {
        let result = [];
        //  Sort dimension properties
        let entries = Object.entries(dimensions).sort((a, b) => a[0].localeCompare(b[0]));
        for (let [name, value] of entries) {
            result.push(`${name}=${value}`);
        }
        return result.join(',');
    }
    /*
        Convert a dimensions string of the form "key=value,..." to a a dimensions object {key: value, ...}
    */
    makeDimensionObject(dimensions) {
        let result = {};
        for (let dimension of dimensions.split(',')) {
            if (dimension) {
                let [key, value] = dimension.split('=');
                result[key] = value;
            }
        }
        return result;
    }
    /*
        Add a point value to the metric the span index with the given timestamp.
        We aggregate aged out points to upper spans as required.
        For queries, a span nominates the desired span, all lower span values are aggregated up.
     */
    addValue(metric, timestamp, point, si, querySpanIndex = undefined) {
        this.assert(metric);
        this.assert(timestamp);
        this.assert(0 <= si && si < metric.spans.length);
        let span = metric.spans[si];
        let interval = span.period / span.samples;
        /* istanbul ignore next */
        let points = span.points || [];
        /*
            Determine if need to recurse and aggregate earlier spans
         */
        let queryRecurse = si < querySpanIndex && si + 1 < metric.spans.length;
        //  Just for safety, should not happen
        /* istanbul ignore next */
        while (points.length > span.samples) {
            points.shift();
        }
        let first = span.end - points.length * interval;
        /*
            Aggregate points. Calculate how many points have aged, or if querying, how many must be aggregated.
        */
        let shift = 0;
        if (points.length) {
            /* istanbul ignore next */
            if (queryRecurse) {
                //  Querying and not yet on the target period, so aggregate all points
                shift = points.length;
            }
            else if (timestamp >= first) {
                //  Count of aged data points
                shift = Math.floor((timestamp - first) / interval) - span.samples;
                if (!queryRecurse && point.count && timestamp >= span.end) {
                    //  Add one more to make room for this point being added
                    shift += 1;
                }
            }
            shift = Math.max(0, Math.min(shift, points.length));
            this.assert(0 <= shift && shift <= points.length);
            /*
                Add points to the next span by shifting aged points and make room.
             */
            for (let i = 0; i < shift; i++) {
                let p = points.shift();
                if (p.count && si + 1 < metric.spans.length) {
                    this.addValue(metric, first, p, si + 1, querySpanIndex);
                }
                first += interval;
            }
        }
        if (queryRecurse) {
            //  Querying and not at the terminal period. Must recurse and aggregate all spans up to the target span.
            this.addValue(metric, timestamp, point, si + 1, querySpanIndex);
            return;
        }
        if (point.count) {
            let index;
            if (points.length == 0) {
                points.push({ count: 0, sum: 0 });
                /*
                    This will set the span.end to the next aligned interval that is >timestamp
                    NOTE: this may mean that span.start is set before now and before recorded time
                 */
                span.end = this.alignTime(span, timestamp + 1);
                first = span.end - interval;
                index = 0;
            }
            else {
                if (timestamp < span.end - span.period) {
                    //  Discard if before the earliest possible point for the span
                    return;
                }
                while (timestamp < first) {
                    points.unshift({ count: 0, sum: 0 });
                    first -= interval;
                }
                while (timestamp >= span.end) {
                    points.push({ count: 0, sum: 0 });
                    span.end += interval;
                }
                index = Math.floor((timestamp - first) / interval);
            }
            this.assert(points.length <= span.samples);
            if (!(0 <= index && index < points.length)) {
                this.assert(0 <= index && index < points.length);
                //  Just in case - should never happen
                /* istanbul ignore next */
                if (index > 0) {
                    index = points.length - 1;
                }
            }
            this.setPoint(span, index, point);
        }
    }
    /*
        Add value to a span and update count, min, max, pValues and sum
        The point "index" defines the point[] to update.
     */
    setPoint(span, index, add) {
        let points = span.points;
        this.assert(0 <= index && index < points.length);
        let point = points[index];
        /* istanbul ignore next */
        if (!point) {
            this.log.error(`Metric null point`, { span, index, add });
            return;
        }
        if (add.count) {
            let value = add.sum / add.count;
            if (point.min == undefined) {
                point.min = value;
            }
            else {
                point.min = Math.min(value, point.min);
            }
            if (point.max == undefined) {
                point.max = value;
            }
            else {
                point.max = Math.max(value, point.max);
            }
        }
        if (this.pResolution) {
            point.pvalues = point.pvalues || [];
            if (add.pvalues) {
                point.pvalues.push(...add.pvalues);
            }
            else {
                point.pvalues.push(add.sum / add.count);
            }
            point.pvalues.splice(0, point.pvalues.length - this.pResolution);
        }
        point.sum += add.sum;
        point.count += add.count;
    }
    /*
        Get list of metrics at a given level. The args: namespace and metrics may be undefined.
        Return {namespaces, metrics, dimensions} as possible.
     */
    async getMetricList(namespace = undefined, metric = undefined, options = { limit: MetricListLimit }) {
        let map = {};
        let owner = options.owner || this.owner;
        let next = options.next;
        let limit = options.limit || MetricListLimit;
        /* istanbul ignore next */
        let chan = options.log == true ? 'info' : 'trace';
        let items, command;
        let count = 0;
        do {
            /* istanbul ignore next */
            ;
            ({ command, items, next } = await this.findMetrics(owner, namespace, metric, limit, next));
            this.log[chan](`Find metrics ${namespace}, ${metric}`, { command, items });
            if (items.length) {
                for (let item of items) {
                    let ns = (map[item.namespace] = map[item.namespace] || {});
                    let met = (ns[item.metric] = ns[item.metric] || []);
                    met.push(item.dimensions);
                }
                count += items.length;
            }
        } while (next && count < limit);
        let result = { namespaces: Object.keys(map) };
        if (namespace && map[namespace]) {
            result.metrics = Object.keys(map[namespace]);
            if (metric) {
                let dimensions = map[namespace][metric];
                if (dimensions) {
                    result.dimensions = [];
                    dimensions = dimensions.sort().filter((v, index, self) => self.indexOf(v) === index);
                    for (let dimension of dimensions) {
                        result.dimensions.push(this.makeDimensionObject(dimension));
                    }
                }
            }
        }
        return result;
    }
    initMetric(owner, namespace, name, dimensions, now) {
        let metric = {
            dimensions,
            metric: name,
            namespace,
            owner,
            spans: [],
            version: Version,
        };
        for (let sdef of this.spans) {
            let span = {
                samples: sdef.samples,
                period: sdef.period,
                end: null,
                points: [],
            };
            //  This will set the span.end to the next aligned interval that is >timestamp
            let interval = span.period / span.samples;
            // span.end = this.alignTime(span, now + 1)
            span.end = this.alignTime(span, now + interval);
            metric.spans.push(span);
        }
        return metric;
    }
    async getMetric(owner, namespace, metric, dimensions, log) {
        let command = new GetItemCommand({
            TableName: this.table,
            Key: {
                [this.primaryKey]: { S: `${this.prefix}#${Version}#${owner}` },
                [this.sortKey]: { S: `${this.prefix}#${namespace}#${metric}#${dimensions}` },
            },
            ConsistentRead: this.consistent,
        });
        let data = await this.client.send(command);
        let result = null;
        if (data && data.Item) {
            let item = unmarshall(data.Item);
            result = this.mapItemFromDB(item);
        }
        if (log == true) {
            let chan = log == true ? 'info' : 'trace';
            this.log[chan](`GetMetric ${namespace}, ${metric} ${dimensions}`, { cmd: command, result });
        }
        return result;
    }
    async findMetrics(owner, namespace, metric, limit, startKey, fields = '') {
        let key = [namespace];
        if (metric) {
            key.push(metric);
        }
        /* istanbul ignore next */
        let start = startKey ? marshall(startKey) : undefined;
        let project = `${this.primaryKey}, ${this.sortKey}`;
        if (fields) {
            project += `, ${fields}`;
        }
        let command = new QueryCommand({
            TableName: this.table,
            ExpressionAttributeNames: {
                '#_0': this.primaryKey,
                '#_1': this.sortKey,
            },
            ExpressionAttributeValues: {
                ':_0': { S: `${this.prefix}#${Version}#${owner}` },
                ':_1': { S: `${this.prefix}#${key.join('#')}` },
            },
            KeyConditionExpression: '#_0 = :_0 and begins_with(#_1, :_1)',
            ConsistentRead: this.consistent,
            Limit: limit,
            ScanIndexForward: true,
            ExclusiveStartKey: start,
            ProjectionExpression: project,
        });
        let result = await this.client.send(command);
        let items = [];
        if (result.Items) {
            for (let i = 0; i < result.Items.length; i++) {
                let item = unmarshall(result.Items[i]);
                items.push(this.mapItemFromDB(item));
            }
        }
        let next = undefined;
        /* istanbul ignore next */
        if (result.LastEvaluatedKey) {
            next = unmarshall(result.LastEvaluatedKey);
        }
        return { items, next, command };
    }
    /*
        Use a sequence number to detect simultaneous updated. If collision, will throw
        a ConditionalCheckFailedException and emit() will then retry.
    */
    async putMetric(item, options) {
        let ConditionExpression, ExpressionAttributeValues;
        let seq;
        if (item.seq != undefined) {
            /* istanbul ignore next */
            seq = item.seq = item.seq || 0;
            /* istanbul ignore next */
            if (item.seq++ >= MaxSeq) {
                item.seq = 0;
            }
            ConditionExpression = `seq = :_0`;
            ExpressionAttributeValues = { ':_0': { N: seq.toString() } };
        }
        else {
            item.seq = 0;
        }
        let mapped = this.mapItemToDB(item);
        let params = {
            TableName: this.table,
            ReturnValues: 'NONE',
            Item: marshall(mapped, { removeUndefinedValues: true }),
            ConditionExpression,
            ExpressionAttributeValues,
        };
        let command = new PutItemCommand(params);
        /* istanbul ignore next */
        let chan = options.log == true ? 'info' : 'trace';
        this.log[chan](`Put metric ${item.namespace}, ${item.metric}`, {
            dimensions: item.dimensions,
            command,
            params,
            item,
        });
        try {
            await this.client.send(command);
            return true;
        }
        catch (err) /* istanbul ignore next */ {
            ;
            (function (err, log) {
                //  SDK V3 puts the code in err.name (Ugh!)
                let code = err.code || err.name;
                if (code == 'ConditionalCheckFailedException') {
                    log.trace(`Update collision`, { err });
                }
                else if (code == 'ProvisionedThroughputExceededException') {
                    log.info(`Provisioned throughput exceeded: ${err.message}`, { err, cmd: command, item });
                }
                else {
                    log.error(`Emit exception code ${err.name} ${err.code} message ${err.message}`, {
                        err,
                        cmd: command,
                        item,
                    });
                    throw err;
                }
                return false;
            })(err, this.log);
        }
    }
    mapItemFromDB(data) {
        let pk = data[this.primaryKey];
        let sk = data[this.sortKey];
        let owner = pk.split('#').pop();
        let [, namespace, metric, dimensions] = sk.split('#');
        let spans;
        if (data.spans) {
            spans = data.spans.map((s) => {
                return {
                    end: s.se,
                    period: s.sp,
                    samples: s.ss,
                    points: s.pt.map((p) => {
                        let point = { count: Number(p.c), sum: Number(p.s) };
                        if (p.x != null) {
                            point.max = Number(p.x);
                        }
                        if (p.m != null) {
                            point.min = Number(p.m);
                        }
                        if (p.v) {
                            point.pvalues = p.v;
                        }
                        return point;
                    }),
                };
            });
        }
        let expires = data[this.expires];
        let seq = data.seq;
        return { dimensions, expires, metric, namespace, owner, seq, spans };
    }
    mapItemToDB(item) {
        let result = {
            [this.primaryKey]: `${this.prefix}#${Version}#${item.owner}`,
            [this.sortKey]: `${this.prefix}#${item.namespace}#${item.metric}#${item.dimensions}`,
            [this.expires]: item.expires,
            spans: item.spans.map((i) => {
                return {
                    se: i.end,
                    sp: i.period,
                    ss: i.samples,
                    pt: i.points.map((point) => {
                        let p = { c: point.count, s: this.round(point.sum) };
                        if (point.max != null) {
                            p.x = this.round(point.max);
                        }
                        if (point.min != null) {
                            p.m = this.round(point.min);
                        }
                        if (point.pvalues) {
                            p.v = point.pvalues;
                        }
                        return p;
                    }),
                };
            }),
            seq: item.seq,
            _source: item._source,
        };
        if (this.type) {
            let [key, model] = Object.entries(this.type)[0];
            result[key] = model;
        }
        return result;
    }
    static freeInstanceByKey(key) {
        delete Instances[key];
    }
    static saveInstance(tags, metrics) {
        let key = JSON.stringify(tags);
        Instances[key] = metrics;
    }
    /*
        Align a timestamp that is >timestamp and rounded up in seconds to be interval aligned
        Note: this may be in the future
     */
    alignTime(span, timestamp) {
        let interval = span.period / span.samples;
        return Math.ceil(timestamp / interval) * interval;
    }
    /* istanbul ignore next */
    assert(c) {
        if (!c && Assert) {
            let msg = { stack: '' };
            if (typeof Error.captureStackTrace === 'function') {
                Error.captureStackTrace(msg);
            }
            else {
                msg.stack = new Error('Assert').stack;
            }
            this.log.error(`Assertion failed`, { stack: msg.stack });
        }
    }
    /* istanbul ignore next */
    info(message, context = {}) {
        console.log('INFO: ' + message, context);
    }
    /* istanbul ignore next */
    error(message, context = {}) {
        console.log('ERROR: ' + message, context);
    }
    /* istanbul ignore next */
    trace(message, context = {}) {
        console.log('TRACE: ' + message, context);
    }
    //  Overcome Javascript number precision issues
    round(n) {
        /* istanbul ignore next */
        if (isNaN(n) || n == null) {
            return 0;
        }
        let places = 16 - n.toFixed(0).length;
        return Number(n.toFixed(places)) - 0;
    }
    /* istanbul ignore next */
    jitter(msecs) {
        return Math.min(10 * 1000, Math.floor(msecs / 2 + msecs * Math.random()));
    }
    /* istanbul ignore next */
    async delay(time) {
        return new Promise(function (resolve, reject) {
            setTimeout(() => resolve(true), time);
        });
    }
}
//  Emulate the SenseLogs logger
/* istanbul ignore next */
class Log {
    senselogs = null;
    logger = null;
    verbose = false;
    constructor(dest) {
        if (dest === true) {
            this.logger = this.defaultLogger;
        }
        else if (dest == 'verbose') {
            this.logger = this.defaultLogger;
            this.verbose = true;
        }
        else if (dest && typeof dest.info == 'function') {
            this.senselogs = dest;
        }
    }
    error(message, context) {
        this.process('error', message, context);
    }
    info(message, context) {
        this.process('info', message, context);
    }
    trace(message, context) {
        this.process('trace', message, context);
    }
    process(chan, message, context) {
        if (this.logger) {
            this.logger(chan, message, context);
        }
        else if (this.senselogs) {
            this.senselogs[chan](message, context);
        }
    }
    /* istanbul ignore next */
    defaultLogger(chan, message, context) {
        if (chan == 'trace' && !this.verbose) {
            //  params.log: true will cause the chan to be changed to 'info'
            return;
        }
        let tag = chan.toUpperCase();
        if (context) {
            try {
                console.log(tag, message, JSON.stringify(context, null, 4));
            }
            catch (err) {
                let buf = ['{'];
                for (let [key, value] of Object.entries(context)) {
                    try {
                        buf.push(`    ${key}: ${JSON.stringify(value, null, 4)}`);
                    }
                    catch (err) {
                        /* continue */
                    }
                }
                buf.push('}');
                console.log(tag, message, buf.join('\n'));
            }
        }
        else {
            console.log(tag, message);
        }
    }
}
